/*
  Copyright (C) 2012  Torsten Brischalle

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#include "lua_func_array.h"

#include "lauxlib.h"
#include "lua_helper.h"

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>

#include <iostream>
#include <sstream>

using namespace std;

//-----------------------------------------------------------------------------

int lua_func_print_array(lua_State *lua)
{
  char delim = ',';
  
  luaL_checktype(lua, 1, LUA_TTABLE);
  
  if (!lua_isnoneornil(lua, 2))
  {
    const char *delim_str = luaL_checkstring(lua, 2);
    if (isprint(delim_str[0]))
      delim = delim_str[0];
  }
  
  int   first;
  
  first = 1;
  for (lua_pushnil(lua); lua_next(lua, 1); lua_pop(lua, 1))
  {
    if (first)
      first = 0;
    else
      cout << delim;
    
    cout << luaL_checknumber(lua, -1);
  }
  
  cout << endl;

  return 0;
}

//-----------------------------------------------------------------------------

int lua_func_make_array(lua_State *lua)
{
  lua_Number start = luaL_checknumber(lua, 1);
  lua_Number end = luaL_checknumber(lua, 2);
  lua_Number step = 1;
  
  if (!lua_isnoneornil(lua, 3))
    step = luaL_checknumber(lua, 3);
  
  if (fabs(end - start) / fabs(step) > 100000)
    lua_throw_error_from_c(lua, "too many values, step width may be to small");
  
  if (end >= start)
    step = fabs(step);
  else
    step = -fabs(step);
    
  lua_Number  val;
  int         i;
  
  lua_newtable(lua);
  for (i = 0; 1; i++)
  {
    val = start + step * i;
    lua_pushnumber(lua, val);
    lua_rawseti(lua, -2, i + 1);
    
    if (((end >= start) && (val >= end)) || ((end < start) && (val <= end)))
      break;
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_reverse_array(lua_State *lua)
{
  if (!lua_istable(lua, 1))
    lua_throw_error_from_c(lua, "parameter is not an array");
  
  int len = lua_objlen(lua, 1), i;
  
  lua_newtable(lua);
  for (i = 0; i < len; i++)
  {
    lua_rawgeti(lua, 1, i + 1);
    lua_rawseti(lua, 2, len - i);
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_mul_array(lua_State *lua)
{
  if (!lua_istable(lua, 1))
    lua_throw_error_from_c(lua, "parameter 1 is not an array");
  
  if (lua_isnumber(lua, 2))
  {
    int      len = lua_objlen(lua, 1), i;
    lua_Number  b = lua_tonumber(lua, 2);
    
    lua_newtable(lua);
    for (i = 1; i <= len; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_pushnumber(lua, a * b);
      lua_rawseti(lua, -2, i);
    }
  }
  else if (lua_istable(lua, 2))
  {
    int  len1 = lua_objlen(lua, 1);
    int  len2 = lua_objlen(lua, 2), i;
    
    if (len1 != len2)
      lua_throw_error_from_c(lua, "the two arrays differs in length");
      
    lua_newtable(lua);
    for (i = 1; i <= len1; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_rawgeti(lua, 2, i);
      lua_Number b = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_pushnumber(lua, a * b);
      lua_rawseti(lua, -2, i);
    }
  }
  else
  {
    lua_throw_error_from_c(lua, "parameter 2 is not an array or a number");
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_div_array(lua_State *lua)
{
  if (!lua_istable(lua, 1))
    lua_throw_error_from_c(lua, "parameter 1 is not an array");
  
  if (lua_isnumber(lua, 2))
  {
    int      len = lua_objlen(lua, 1), i;
    lua_Number  b = lua_tonumber(lua, 2);
    
    if (b == 0)
      lua_throw_error_from_c(lua, "division by zero");
      
    lua_newtable(lua);
    for (i = 1; i <= len; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_pushnumber(lua, a / b);
      lua_rawseti(lua, -2, i);
    }
  }
  else if (lua_istable(lua, 2))
  {
    int  len1 = lua_objlen(lua, 1);
    int  len2 = lua_objlen(lua, 2), i;
    
    if (len1 != len2)
      lua_throw_error_from_c(lua, "the two arrays differs in length");
      
    lua_newtable(lua);
    for (i = 1; i <= len1; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_rawgeti(lua, 2, i);
      lua_Number b = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      if (b == 0)
        lua_throw_error_from_c(lua, "division by zero");
      
      lua_pushnumber(lua, a / b);
      lua_rawseti(lua, -2, i);
    }
  }
  else
  {
    lua_throw_error_from_c(lua, "parameter 2 is not an array or a number");
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_sub_array(lua_State *lua)
{
  if (!lua_istable(lua, 1))
    lua_throw_error_from_c(lua, "parameter 1 is not an array");
  
  if (lua_isnumber(lua, 2))
  {
    int      len = lua_objlen(lua, 1), i;
    lua_Number  b = lua_tonumber(lua, 2);
    
    lua_newtable(lua);
    for (i = 1; i <= len; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_pushnumber(lua, a - b);
      lua_rawseti(lua, -2, i);
    }
  }
  else if (lua_istable(lua, 2))
  {
    int  len1 = lua_objlen(lua, 1);
    int  len2 = lua_objlen(lua, 2), i;
    
    if (len1 != len2)
      lua_throw_error_from_c(lua, "the two arrays differs in length");
      
    lua_newtable(lua);
    for (i = 1; i <= len1; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_rawgeti(lua, 2, i);
      lua_Number b = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_pushnumber(lua, a - b);
      lua_rawseti(lua, -2, i);
    }
  }
  else
  {
    lua_throw_error_from_c(lua, "parameter 2 is not an array or a number");
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_add_array(lua_State *lua)
{
  if (!lua_istable(lua, 1))
    lua_throw_error_from_c(lua, "parameter 1 is not an array");
  
  if (lua_isnumber(lua, 2))
  {
    int      len = lua_objlen(lua, 1), i;
    lua_Number  b = lua_tonumber(lua, 2);
    
    lua_newtable(lua);
    for (i = 1; i <= len; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_pushnumber(lua, a + b);
      lua_rawseti(lua, -2, i);
    }
  }
  else if (lua_istable(lua, 2))
  {
    int  len1 = lua_objlen(lua, 1);
    int  len2 = lua_objlen(lua, 2), i;
    
    if (len1 != len2)
      lua_throw_error_from_c(lua, "the two arrays differs in length");
      
    lua_newtable(lua);
    for (i = 1; i <= len1; i++)
    {
      lua_rawgeti(lua, 1, i);
      lua_Number a = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_rawgeti(lua, 2, i);
      lua_Number b = lua_tonumber(lua, -1);
      lua_pop(lua, 1);
      
      lua_pushnumber(lua, a + b);
      lua_rawseti(lua, -2, i);
    }
  }
  else
  {
    lua_throw_error_from_c(lua, "parameter 2 is not an array or a number");
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_extract_array(lua_State *lua)
{
  if (!lua_istable(lua, 1))
    lua_throw_error_from_c(lua, "parameter 1 is not an array");
  
  int len = (int)lua_objlen(lua, 1);
  
  int start_idx = (int)luaL_checknumber(lua, 2);
  
  if (start_idx < 1)
    lua_throw_error_from_c(lua, "start index is less than 1");
    
  if (start_idx > len)
    lua_throw_error_from_c(lua, "start index is greater than array length");
  
  int step_size = 1;
  if (!lua_isnoneornil(lua, 3))
    step_size = (int)luaL_checknumber(lua, 3);
    
  int count = 0x7FFFFFFF;
  if (!lua_isnoneornil(lua, 4))
    count = (int)luaL_checknumber(lua, 4);
  
  int i, c;
  
  lua_newtable(lua);
  
  if (count > 0)
  {
    c = 1;
    for (i = start_idx; (i >= 1) && (i <= len); i += step_size)
    {
      lua_rawgeti(lua, 1, i);
      lua_rawseti(lua, -2, c);
      
      c++;
      if (c > count)
        break;
    }
  }
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_array_min(lua_State *lua)
{
  lua_Number  res = 0;
  int         first = 1;
  
  luaL_checktype(lua, 1, LUA_TTABLE);
  
  for (lua_pushnil(lua); lua_next(lua, 1); lua_pop(lua, 1))
  {
    if (first)
    {
      res = luaL_checknumber(lua, -1);
      first = 0;
    }
    else
    {
      lua_Number n = luaL_checknumber(lua, -1);
      
      if (n < res)
        res = n;
    }
  }
  
  lua_pushnumber(lua, res);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_array_max(lua_State *lua)
{
  lua_Number  res = 0;
  int         first = 1;
  
  luaL_checktype(lua, 1, LUA_TTABLE);
  
  for (lua_pushnil(lua); lua_next(lua, 1); lua_pop(lua, 1))
  {
    if (first)
    {
      res = luaL_checknumber(lua, -1);
      first = 0;
    }
    else
    {
      lua_Number n = luaL_checknumber(lua, -1);
      
      if (n > res)
        res = n;
    }
  }
  
  lua_pushnumber(lua, res);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_array_to_string(lua_State *lua)
{
  char delim = ',';
  
  luaL_checktype(lua, 1, LUA_TTABLE);
  
  if (!lua_isnoneornil(lua, 2))
  {
    const char *delim_str = luaL_checkstring(lua, 2);
    if (isprint(delim_str[0]))
      delim = delim_str[0];
  }
  
  int           first;
  ostringstream ss;

  first = 1;
  for (lua_pushnil(lua); lua_next(lua, 1); lua_pop(lua, 1))
  {
    if (first)
      first = 0;
    else
      ss << delim;
    
    ss << luaL_checknumber(lua, -1);
  }
  
  lua_pushstring(lua, ss.str().c_str());
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_array_to_number(lua_State *lua)
{
  luaL_checktype(lua, 1, LUA_TTABLE);
  
  lua_newtable(lua);
  
  int   len = lua_objlen(lua, 1);
  int   idx;
  
  for (idx = 1; idx <= len; idx++)
  {
    lua_rawgeti(lua, 1, idx);
    
    const char *item = luaL_checkstring(lua, -1);
    lua_Number val = 0;
    
    if (item)
      val = strtod(item, 0);
    
    lua_pop(lua, 1);
    
    lua_pushnumber(lua, val);
    lua_rawseti(lua, 2, idx);
  }
  
  return 1;
}

//-----------------------------------------------------------------------------
