/*
  Copyright (C) 2012  Torsten Brischalle

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/
*/

#include "lua_func_bit.h"

#include "lauxlib.h"

#include <stdint.h>

#include "lua_helper.h"

//-----------------------------------------------------------------------------

int lua_func_bit_or(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  uint32_t op2 = (uint32_t)luaL_checknumber(lua, 2);
  
  lua_pushnumber(lua, op1 | op2);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_and(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  uint32_t op2 = (uint32_t)luaL_checknumber(lua, 2);
  
  lua_pushnumber(lua, op1 & op2);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_xor(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  uint32_t op2 = (uint32_t)luaL_checknumber(lua, 2);
  
  lua_pushnumber(lua, op1 ^ op2);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_nand(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  uint32_t op2 = (uint32_t)luaL_checknumber(lua, 2);
  
  lua_pushnumber(lua, ~(op1 & op2));
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_nor(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  uint32_t op2 = (uint32_t)luaL_checknumber(lua, 2);
  
  lua_pushnumber(lua, ~(op1 | op2));
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_xnor(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  uint32_t op2 = (uint32_t)luaL_checknumber(lua, 2);
  
  lua_pushnumber(lua, ~(op1 ^ op2));
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_not(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  
  lua_pushnumber(lua, ~op1);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_get(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  int      bit_num = (int)luaL_checknumber(lua, 2);
  
  if ((bit_num < 0) || (bit_num > 31))
    lua_throw_error_from_c(lua, "invalid bit number (valid range: 0 .. 31)");
  
  lua_pushnumber(lua, (op1 & (1 << bit_num)) ? 1 : 0);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_set(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  int      bit_num = (int)luaL_checknumber(lua, 2);
  
  if ((bit_num < 0) || (bit_num > 31))
    lua_throw_error_from_c(lua, "invalid bit number (valid range: 0 .. 31)");
  
  lua_pushnumber(lua, op1 | (1 << bit_num));
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_clear(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  int      bit_num = (int)luaL_checknumber(lua, 2);
  
  if ((bit_num < 0) || (bit_num > 31))
    lua_throw_error_from_c(lua, "invalid bit number (valid range: 0 .. 31)");
  
  lua_pushnumber(lua, op1 & (~(1 << bit_num)));
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_toggle(lua_State *lua)
{
  uint32_t op1 = (uint32_t)luaL_checknumber(lua, 1);
  int      bit_num = (int)luaL_checknumber(lua, 2);
  
  if ((bit_num < 0) || (bit_num > 31))
    lua_throw_error_from_c(lua, "invalid bit number (valid range: 0 .. 31)");
  
  lua_pushnumber(lua, op1 ^ (1 << bit_num));
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_shr(lua_State *lua)
{
  uint32_t val = (uint32_t)luaL_checknumber(lua, 1);
  int      cnt = (int)luaL_checknumber(lua, 2);
  
  if (cnt < 0)
    lua_throw_error_from_c(lua, "parameter 2 must be greater than or equal to 0");
  
  lua_pushnumber(lua, val >> cnt);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_shl(lua_State *lua)
{
  uint32_t val = (uint32_t)luaL_checknumber(lua, 1);
  int      cnt = (int)luaL_checknumber(lua, 2);
  
  if (cnt < 0)
    lua_throw_error_from_c(lua, "parameter 2 must be greater than or equal to 0");
  
  lua_pushnumber(lua, val << cnt);
  
  return 1;
}

//-----------------------------------------------------------------------------

int lua_func_bit_reverse(lua_State *lua)
{
	uint32_t val = (uint32_t)luaL_checknumber(lua, 1);
	uint32_t res = 0;
	int      cnt = (int)luaL_checknumber(lua, 2);
	int      i;

	if ((cnt < 1) || (cnt > 32))
		lua_throw_error_from_c(lua, "Invalid number of bits. (Valid range: 1 .. 32)");

	for (i = 0; i < cnt; i++)
	{
		res <<= 1;
		res |= (val & 1);
		val >>= 1;
	}

	lua_pushnumber(lua, res);

	return 1;
}

//-----------------------------------------------------------------------------
